
/*
 *
 *  A set of routines to read and write integers and IEEE
 *  floating-point numbers taking into account the byte order.
 *
 *  Based on the C++ advanced I/O library written by Oleg Kiselyov.
 *
 *  Copyright 2005-2008 Ivan Raikov and the Georgia Institute of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * A full copy of the GPL license can be found at
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>

#include "endian_lowio.h"
#include "floatformat.h"

#define ENDIAN_LOWIO_ERROR_MSG_LEN 512
char endian_lowio_error_message[ENDIAN_LOWIO_ERROR_MSG_LEN+1];
int endian_lowio_error_status = 0;
char *endian_lowio_error_string;

void endian_lowio_error(const char *descr)
{
     perror(descr);
     strncpy (endian_lowio_error_message, descr, ENDIAN_LOWIO_ERROR_MSG_LEN);
     /* make sure the string is null-terminated */
     endian_lowio_error_message[ENDIAN_LOWIO_ERROR_MSG_LEN]=0;
     endian_lowio_error_string = endian_lowio_error_message;
     endian_lowio_error_status = EIO;
}

void endian_lowio_eof(const char *descr)
{
     snprintf (endian_lowio_error_message, ENDIAN_LOWIO_ERROR_MSG_LEN,
	       "%s: end of file reached", descr);
     /* make sure the string is null-terminated */
     endian_lowio_error_message[ENDIAN_LOWIO_ERROR_MSG_LEN]=0;
     endian_lowio_error_string = endian_lowio_error_message;
     endian_lowio_error_status = EIO;
}

int endian_lowio_assert_append(const char * file_name)
{
     FILE *f;

     if ((f = fopen (file_name, "a+")) != NULL)
	  return fileno(f);

     snprintf (endian_lowio_error_message, ENDIAN_LOWIO_ERROR_MSG_LEN,
	       "failed to open file [%s] for appending", file_name);
     /* make sure the string is null-terminated */
     endian_lowio_error_message[ENDIAN_LOWIO_ERROR_MSG_LEN]=0;
     endian_lowio_error_string = endian_lowio_error_message;
     endian_lowio_error_status = EIO;

     return -1;
}

int endian_lowio_assert_create(const char * file_name)
{
     FILE *f;

     if ((f = fopen (file_name, "w+")) != NULL)
	  return fileno(f);

     snprintf (endian_lowio_error_message, ENDIAN_LOWIO_ERROR_MSG_LEN,
	       "failed to create file [%s]", file_name);
     /* make sure the string is null-terminated */
     endian_lowio_error_message[ENDIAN_LOWIO_ERROR_MSG_LEN]=0;
     endian_lowio_error_string = endian_lowio_error_message;
     endian_lowio_error_status = EIO;

     return -1;
}

// Read a CHAR (1 byte) integer 
unsigned char endian_lowio_read_int1(int fd, byte_order_t b)
{
     ssize_t s;
     char c1;
     if ( (s = read(fd,&c1,1)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_read_int1");
     }
     if (s == 0) endian_lowio_eof ("endian_lowio_read_int1");

     return c1;
}

// Read a SHORT (2 bytes) integer in the specified endian order
unsigned short int endian_lowio_read_int2(int fd, byte_order_t b)
{
     ssize_t s;
     unsigned char c[2];

     if ( (s = read(fd,&c,2)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_read_int2");
     }
     if (s == 0) endian_lowio_eof ("endian_lowio_read_int2");
     
     if( endian_lowio_is_MSB(b) )
	  return ((unsigned char)c[0] << 8) | (unsigned char)c[1];
     else
	  return ((unsigned char)c[1] << 8) | (unsigned char)c[0];
}

// Read a LONG (4 bytes) integer in the specified endian order
unsigned long int endian_lowio_read_int4(int fd, byte_order_t b)
{
     ssize_t s;
     unsigned char c[4];

     if ( (s = read(fd,&c,4)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_read_int4");
     }
     if (s == 0) endian_lowio_eof ("endian_lowio_read_int4");

     if ( endian_lowio_is_MSB(b) )
	  return ((unsigned char)c[0] << 24) | ((unsigned char)c[1] << 16) |
	       ((unsigned char)c[2] << 8) | (unsigned char)c[3];
     else
	  return ((unsigned char)c[3] << 24) | ((unsigned char)c[2] << 16) |
	       ((unsigned char)c[1] << 8) | (unsigned char)c[0];
}


// Read a LONG LONG (8 bytes) integer in the specified endian order
unsigned long long endian_lowio_read_int8 (int fd, byte_order_t b)
{
     ssize_t s;
     unsigned char c[8];

     if ( (s = read(fd,&c,8)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_read_int8");
     }
     if (s == 0) endian_lowio_eof ("endian_lowio_read_int8");

     if ( endian_lowio_is_MSB(b) )
     {
	  // c1 is most-significant byte
	  return ((unsigned long long)c[0] << 56) | 
	         ((unsigned long long)c[1] << 48) | 
	         ((unsigned long long)c[2] << 40) | 
	         ((unsigned long long)c[3] << 32) |
	         ((unsigned long long)c[4] << 24) | 
	         ((unsigned long long)c[5] << 16) |
	         ((unsigned long long)c[6] << 8)  | 
	         (unsigned long long)c[7];
     }
     else
     {
	  // c8 is most-significant byte
	  return ((unsigned long long)c[7] << 56) | 
	         ((unsigned long long)c[6] << 48) | 
	         ((unsigned long long)c[5] << 40) | 
	         ((unsigned long long)c[4] << 32) |
	         ((unsigned long long)c[3] << 24) | 
	         ((unsigned long long)c[2] << 16) |
	         ((unsigned long long)c[1] << 8)  | 
	         (unsigned long long)c[0];
     }
}


// Read a single IEEE float in the specified endian order
void endian_lowio_read_ieee_float32 (int fd, byte_order_t b, int *nan, double *f)
{
     ssize_t s;
     char c[4];
     double d;
     const struct floatformat *fmt;
     unsigned long exponent;

     d = 0.0;

     if ( (s = read(fd,&c,4)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_read_ieee_float32");
     }
     if (s == 0) endian_lowio_eof ("endian_lowio_read_ieee_float32");

     
     if ( endian_lowio_is_MSB(b) )
     {
	  fmt = &floatformat_ieee_single_big;
     }
     else
     {
	  fmt = &floatformat_ieee_single_little;
     }

     exponent = get_field ((unsigned char *)c, fmt->byteorder, fmt->totalsize,
			   fmt->exp_start, fmt->exp_len);

     if (exponent == fmt->exp_nan)
     {
	  *nan = 1;
     }
     else
     {
	  *nan = 0;
	  floatformat_to_double(fmt, c, &d);
     }
     *f = d;
}

// Read a double IEEE float in the specified endian order
void endian_lowio_read_ieee_float64 (int fd, byte_order_t b, int *nan, double *d)
{
     ssize_t s;
     char c[8];
     const struct floatformat *fmt;
     unsigned long exponent;

     *d = 0.0;

     if ( (s = read(fd,&c,8)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_read_ieee_float64");
     }
     if (s == 0) endian_lowio_eof ("endian_lowio_read_ieee_float64");

     if ( endian_lowio_is_MSB(b) )
     {
	  fmt = &floatformat_ieee_double_big;
     }
     else
     {
	  fmt = &floatformat_ieee_double_little;
     }

     exponent = get_field ((unsigned char *)c, fmt->byteorder, fmt->totalsize,
			   fmt->exp_start, fmt->exp_len);
     if (exponent == fmt->exp_nan)
     {
	  *nan = 1;
     }
     else
     {
	  *nan = 0;
	  floatformat_to_double(fmt, c, d);
     }

}


int write_intn (int fd, const unsigned int size, const char *d, const char *desc)
{
     ssize_t s,n;

     n = s = 0;
     while (n < size)
     {
	  if ( (s = write(fd,(const void *)(d+n),size-n)) == -1 )
	  {
	       endian_lowio_error(desc);
	       return -1;
	  }
	  n += s;
     }
     return 0;
}


// Write  a byte in the specified endian order
void endian_lowio_write_int1 (int fd, byte_order_t b, const char item)
{
     ssize_t s;
     if ( (s = write(fd,&item,1)) == -1 )
     {
	  endian_lowio_error ("endian_lowio_write_int1");
     }
}

// Write  a short integer in the specified endian order
void endian_lowio_write_int2 (int fd, byte_order_t b, const unsigned short item)
{
     ssize_t s;
     unsigned char c[2];

     if ( endian_lowio_is_MSB(b) )
     {
	  c[0] = (item>>8) & 0xff;
	  c[1] = item & 0xff;
     }
     else
     {
	  c[0] = item & 0xff;
	  c[1] = (item>>8) & 0xff;
     }

     write_intn(fd,sizeof(c),c,"endian_lowio_write_int2");
}


// Write out a long integer in the specified endian order
void endian_lowio_write_int4 (int fd, byte_order_t b, const unsigned long item)
{
     ssize_t s;
     char c[4]; int i, k;
     unsigned int t = item;
     
     if ( endian_lowio_is_MSB(b) )
     {
	  k = 24;
	  for(i=0; i < 4; i++)
	  {
	       c[i] = (t >> k) & 0xff;
	       k -= 8;
	  }
     }
     else
     {
	  for(i=0; i < 4; i++)
	  {
	       c[i] = t & 0xff;
	       t >>= 8;
	  }
     }

     write_intn(fd,sizeof(c),c,"endian_lowio_write_int4");
}

// Write out a long long integer in the specified endian order
void endian_lowio_write_int8 (int fd, byte_order_t b, const unsigned long long item)
{
     ssize_t s;
     char c[8]; int i, k;
     unsigned int t = item;
     
     if ( endian_lowio_is_MSB(b) )
     {
	  k = 56;
	  for(i=0; i < 8; i++)
	  {
	       c[i] = (t >> k) & 0xff;
	       k -= 8;
	  }
     }
     else
     {
	  for(i=0; i < 8; i++)
	  {
	       c[i] = t & 0xff;
	       t >>= 8;
	  }
     }

     write_intn(fd,sizeof(c),c,"endian_lowio_write_int8");
}

// Write out a single IEEE float in the specified endian order
void  endian_lowio_write_ieee_float32 (int fd, byte_order_t b, const float n)
{
     ssize_t s;
     char c[4];
     double d;

     d = n;

     if ( endian_lowio_is_MSB(b) )
     {
	  double_to_floatformat (&floatformat_ieee_single_big, &d, c);
     }
     else
     {
	  double_to_floatformat (&floatformat_ieee_single_little, &d, c);
     }
     

     write_intn(fd,sizeof(4),c,"endian_lowio_write_ieee_float32");
}

// Write out a double IEEE float in the specified endian order
void  endian_lowio_write_ieee_float64 (int fd, byte_order_t b, const double n)
{
     ssize_t s;
     char c[8];
     double d;

     d = n;

     if ( endian_lowio_is_MSB(b) )
     {
	  double_to_floatformat (&floatformat_ieee_double_big, &d, c);
     }
     else
     {
	  double_to_floatformat (&floatformat_ieee_double_little, &d, c);
     }

     write_intn(fd,sizeof(c),c,"endian_lowio_write_ieee_float64");
}
